from abc import ABC, abstractmethod


class TrackObserver(ABC):
    """Abstract base class for all TrackObservers."""

    name = "TrackObserver"

    def get_name(self):
        return self.name

    @abstractmethod
    def track_started(self, track):  # pragma: no cover
        pass

    @abstractmethod
    def track_finished(self, track):  # pragma: no cover
        pass



import configargparse


class Options(object):
    """Contain all hardcoded and loaded from configargparse options."""

    """How many seconds the main daemon loop sleeps."""
    sleepSeconds = 1

    """Default socket of 2 minutes, to prevent endless hangs on HTTP requests."""
    socketDefaultTimeout = 120

    def __init__(self):
        """Configure configargparse."""
        self.__args = configargparse.ArgParser(
            default_config_files=["/etc/nowplaying/conf.d/*.conf", "~/.nowplayingrc"]
        )
        self.__args.add_argument(
            "-b",
            "--saemubox-ip",
            dest="saemubox_ip",
            help="IP address of SAEMUBOX",
            default="",
        )
        # TODO v3 remove this option
        self.__args.add_argument(
            "-m",
            "--icecast-base",
            dest="icecastBase",
            help="Icecast base URL",
            default="http://icecast.example.org:8000/admin/",
        )
        # TODO v3 remove this option
        self.__args.add_argument(
            "--icecast-password", dest="icecastPassword", help="Icecast Password"
        )
        self.__args.add_argument(
            "-i",
            "--icecast",
            action="append",
            help="""Icecast endpoints, allowed multiple times. nowplaying will send metadata updates to each
            of the configured endpoints. Specify complete connection data like username and password in the
            URLs e.g. 'http://source:changeme@icecast.example.org:8000/admin/metadata.xsl?mount=/radio'.""",
            default=[],
        )
        self.__args.add_argument(
            "-d",
            "--dab",
            action="append",
            help="DAB audio companion base URL, allowed multiple times (ie. http://dab.example.org:8080)",
            default=[],
        )
        # TODO v3 remove when stable
        self.__args.add_argument(
            "--dab-send-dls",
            type=bool,
            nargs="?",
            dest="dab_send_dls",
            help="Send artist/title to DAB companions dls endpoint (default: True)",
            default=True,
        )
        self.__args.add_argument(
            "-s",
            "--show",
            dest="currentShowUrl",
            help="Current Show URL e.g. 'https://libretime.int.example.org/api/live-info-v2/format/json'",
        )
        self.__args.add_argument(
            "--input-file",
            dest="inputFile",
            help="XML 'now-playing' input file location",
            default="/home/endlosplayer/Eingang/now-playing.xml",
        )
        self.__args.add_argument(
            "--xml-output",
            dest="tickerOutputFile",
            help="ticker XML output format",
            default="/var/www/localhost/htdocs/songticker/0.9.3/current.xml",
        )
        self.__args.add_argument(
            "--instrumentation-otlp-enable",
            type=bool,
            nargs="?",
            const=True,
            dest="otlp_enable",
            help="Enable OpenTelemetry Protocol (OTLP) exporter (default: False)",
            default=False,
            env_var="NOWPLAYING_INSTRUMENTATION_OTLP_ENABLE",
        )
        self.__args.add_argument(
            "--debug",
            type=bool,
            nargs="?",
            const=True,
            dest="debug",
            help="Show debug messages",
            default=False,
        )

    def parse_known_args(self):
        """Parse known args with configargparse."""
        self.__args.parse_known_args(namespace=self)



import logging
from typing import Optional, Tuple
from urllib.parse import parse_qs, urlparse

logger = logging.getLogger(__name__)


def parse_icecast_url(
    url: str,
) -> Tuple[str, Optional[str], Optional[str], Optional[str]]:
    """Parse an Icecast URL into the componets relevant to the :class:`IcecastTrackObserver`.

    Use it to grab the username, password, and mountpoint from the URL.

    >>> parse_icecast_url("http://user:password@localhost:80/?mount=foo.mp3")
    ('http://localhost:80/', 'user', 'password', 'foo.mp3')

    It returns None values for missing parts.

    >>> parse_icecast_url("http://localhost/")
    ('http://localhost:80/', None, None, None)

    It supports https URLs.

    >>> parse_icecast_url("https://localhost/")
    ('https://localhost:443/', None, None, None)

    Args:
        url (str): The Icecast URL to parse.
    Returns:
        Tuple[str, Optional[str], Optional[str], Optional[str]]: The URL, username, password, and mountpoint.
    """
    parsed = urlparse(url)
    port = parsed.port or parsed.scheme == "https" and 443 or 80
    url = parsed._replace(query="", netloc=f"{parsed.hostname}:{port}").geturl()
    username = parsed.username
    password = parsed.password
    mount = None
    try:
        mount = parse_qs(parsed.query)["mount"][0]
    except KeyError:
        logger.warning("Missing mount parameter in URL %s" % url)
    return (url, username, password, mount)



import datetime
import logging
import logging.handlers
import uuid

import pytz

logger = logging.getLogger(__name__)

DEFAULT_ARTIST = "Radio Bern"
DEFAULT_TITLE = "Livestream"


class TrackError(Exception):
    """Track related exception."""

    pass


class Track:
    """Track object which has a start and end time and a related show."""

    def __init__(self):
        self.artist = None

        self.title = None

        self.album = None

        self.track = 1

        # The track's global unique identifier
        self.uuid = str(uuid.uuid4())

        # Get current datetime object in UTC timezone
        now = datetime.datetime.now(pytz.timezone("UTC"))

        # The show's start time, initially set to now
        self.starttime = now

        # The show's end time, initially set to to now
        self.endtime = now

    def set_artist(self, artist):
        self.artist = artist

    def set_title(self, title):
        self.title = title

    def set_album(self, album):
        self.album = album

    def set_track(self, track):
        if track < 0:
            raise TrackError("track number has to be a positive integer")

        self.track = track

    def set_starttime(self, starttime):
        if not isinstance(starttime, datetime.datetime):
            raise TrackError("starttime has to be a datatime object")

        # The track's start time as a datetime object
        self.starttime = starttime

    def set_endtime(self, endtime):
        if not isinstance(endtime, datetime.datetime):
            raise TrackError("endtime has to be a datatime object")

        # The track's end time as a datetime object
        self.endtime = endtime

    def set_duration(self, seconds):
        self.endtime = self.starttime + datetime.timedelta(seconds=int(seconds))

    def set_show(self, show):
        # if not isinstance(show, show.Show):
        #    raise TrackError('show has to be a Show object')

        # The show which the track is related to
        self.show = show

    def get_duration(self):
        return self.endtime - self.starttime

    def has_default_artist(self):
        if self.artist == DEFAULT_ARTIST:
            return True

        return False

    def has_default_title(self):
        if self.title == DEFAULT_TITLE:
            return True

        return False

    def __str__(self):
        return "Track '%s' - '%s', start: '%s', end: '%s', uid: %s" % (
            self.artist,
            self.title,
            self.starttime,
            self.endtime,
            self.uuid,
        )



"""Entrypoint for nowplaing module.

This module is the entrypoint for the nowplaying module.

You can run it via `python3 -mnowplaying` or after installation using the `nowplaying` command.
"""
from .main import NowPlaying


def main():
    NowPlaying().run()


if __name__ == "__main__":  # pragma: no cover
    main()



import logging
import logging.handlers
import os
import time
import warnings
import xml.dom.minidom
from abc import ABC, abstractmethod

import isodate
import pytz

from ..show import client
from ..show.show import Show
from ..track.track import DEFAULT_ARTIST, DEFAULT_TITLE, Track

logger = logging.getLogger(__name__)


class InputObserver(ABC):
    """Abstract base for all InputObservers."""

    _SHOW_NAME_KLANGBECKEN = "Klangbecken"
    _SHOW_URL_KLANGBECKEN = "http://www.rabe.ch/sendungen/musik/klangbecken.html"

    def __init__(self, current_show_url):
        self.current_show_url = current_show_url

        self.first_run = True
        self.previous_saemubox_id = None
        self.show = None
        self.showclient = client.ShowClient(current_show_url)
        self.show = self.showclient.get_show_info()

        self.previous_show_uuid = None

        self.track_handler = None

    def add_track_handler(self, track_handler):
        self.track_handler = track_handler

    def update(self, saemubox_id):
        if self.handle_id(saemubox_id):
            self.handle()

    @abstractmethod
    def handle_id(self, saemubox_id):  # pragma: no coverage
        pass

    @abstractmethod
    def handle(self):  # pragma: no coverage
        pass


class KlangbeckenInputObserver(InputObserver):
    """Observe cases where Sämu Box says Klangbecken is running and we can consume now-playing.xml input."""

    def __init__(self, current_show_url, input_file):  # pragma: no coverage
        # TODO test once input file is replaced with api
        warnings.warn(
            "The now-playing.xml format from Loopy/Klangbecken will be replaced in the future",
            PendingDeprecationWarning,
        )
        InputObserver.__init__(self, current_show_url)

        self.input_file = input_file
        self.last_modify_time = os.stat(self.input_file).st_mtime

        self.track = None

    def handle_id(self, saemubox_id):
        # only handle Klangbecken output
        if saemubox_id == 1:
            return True

        return False

    def handle(self):
        # @TODO: replace the stat method with inotify
        modify_time = os.stat(self.input_file).st_mtime

        # @TODO: Need to check if we have a stale file and send default
        #        track infos in this case. This might happend if loopy
        #        went out for lunch...
        #        pseudo code: now > modify_time + self.track.get_duration()

        if self.first_run or modify_time > self.last_modify_time:
            logger.info("Now playing file changed")

            self.show = self.showclient.get_show_info()
            self.last_modify_time = modify_time

            logger.info("First run: %s" % self.first_run)

            if not self.first_run:  # pragma: no coverage
                # TODO test once i don't have to care about mtime/inotify because it's an api
                logger.info("calling track_finished")
                self.track_handler.track_finished(self.track)

            self.track = self.get_track_info()

            # Klangbecken acts as a failover and last resort input, if other
            # active inputs are silent or have problems.
            # Therefore the show's name should always be Klangbecken, regardless
            # of what loopy thinks.
            if self.show.name != self._SHOW_NAME_KLANGBECKEN:
                logger.info(
                    "Klangbecken Input active, overriding current show '%s' with '%s'"
                    % (self.show.name, self._SHOW_NAME_KLANGBECKEN)
                )

                self.show = Show()
                self.show.set_name(self._SHOW_NAME_KLANGBECKEN)
                self.show.set_url(self._SHOW_URL_KLANGBECKEN)

                # Set the show's end time to the one of the track, as we have
                # no idea for how long the Klangbecken input will be active.
                # The show's start time is initially set to now.
                self.show.set_endtime(self.track.endtime)

            self.track.set_show(self.show)

            self.track_handler.track_started(self.track)

            self.first_run = False

    def get_track_info(self):
        dom = xml.dom.minidom.parse(self.input_file)

        # default track info
        track_info = {
            "artist": DEFAULT_ARTIST,
            "title": DEFAULT_TITLE,
            "album": "",
            "track": "",
            "time": "",
        }

        song = dom.getElementsByTagName("song")

        if len(song) == 0 or song[0].hasChildNodes() is False:  # pragma: no coverage
            # TODO replace with non generic exception and test
            raise Exception("No <song> tag found")

        song = song[0]

        for name in list(track_info.keys()):
            elements = song.getElementsByTagName(name)

            if len(elements) == 0:  # pragma: no coverage
                # TODO replace with non generic exception and test
                raise Exception("No <%s> tag found" % name)
            elif elements[0].hasChildNodes():
                element_data = elements[0].firstChild.data.strip()

                if element_data != "":
                    track_info[name] = element_data
                else:  # pragma: no coverage
                    logger.info("Element %s has empty value, ignoring" % name)

        if not song.hasAttribute("timestamp"):  # pragma: no coverage
            # TODO replace with non generic exception and test
            raise Exception("Song timestamp attribute is missing")

        # set the start time and append the missing UTC offset
        # @TODO: The UTC offset should be provided by the now playing XML
        #        generated by Thomas
        # ex.: 2012-05-15T09:47:07+02:00
        track_info["start_timestamp"] = song.getAttribute("timestamp") + time.strftime(
            "%z"
        )

        current_track = Track()

        current_track.set_artist(track_info["artist"])
        current_track.set_title(track_info["title"])
        current_track.set_album(track_info["album"])

        # Store as UTC datetime object
        current_track.set_starttime(
            isodate.parse_datetime(track_info["start_timestamp"]).astimezone(
                pytz.timezone("UTC")
            )
        )

        current_track.set_duration(track_info["time"])

        return current_track


class NonKlangbeckenInputObserver(InputObserver):
    """Observer for input that doesn't originate from klangbecken and therefore misses the track information.

    Uses the show's name instead of the actual track infos
    """

    def handle_id(self, saemubox_id):

        if saemubox_id != self.previous_saemubox_id:
            # If sämubox changes, force a show update, this acts as
            # a self-healing measurement in case the show web service provides
            # nonsense ;)
            self.show = self.showclient.get_show_info(True)

        self.previous_saemubox_id = saemubox_id

        # only handle non-Klangbecken
        if saemubox_id != 1:
            return True

        return False

    def handle(self):
        self.show = self.showclient.get_show_info()

        # only handle if a new show has started
        if self.show.uuid != self.previous_show_uuid:
            logger.info("Show changed")
            self.track_handler.track_started(self.get_track_info())
            self.previous_show_uuid = self.show.uuid

    def get_track_info(self):
        current_track = Track()

        current_track.set_artist(DEFAULT_ARTIST)
        current_track.set_title(DEFAULT_TITLE)

        # Set the track's start/end time to the start/end time of the show
        current_track.set_starttime(self.show.starttime)
        current_track.set_endtime(self.show.endtime)

        current_track.set_show(self.show)

        return current_track



import logging
import logging.handlers

from .observer import InputObserver

logger = logging.getLogger(__name__)


class InputHandler:
    """Inform all registered input-event observers about an input status.

    This is the subject of the classical observer pattern.
    """

    def __init__(self):
        self._observers = []

    def register_observer(self, observer: InputObserver):
        logger.info("Registering InputObserver '%s'" % observer.__class__.__name__)
        self._observers.append(observer)

    def remove_observer(self, observer: InputObserver):
        self._observers.remove(observer)

    def update(self, saemubox_id: int):
        for observer in self._observers:
            logger.debug("Sending update event to observer %s" % observer.__class__)

            try:
                observer.update(saemubox_id)
            except Exception as e:  # pragma: no cover
                # TODO test once replaced with non generic exception
                logger.error("InputObserver (%s): %s" % (observer.__class__, e))
                logger.exception(e)



import logging
from typing import Optional

import requests

from ...util import parse_icecast_url
from ..track import Track
from .base import TrackObserver

logger = logging.getLogger(__name__)


class IcecastTrackObserver(TrackObserver):
    """Update track metadata on an icecast mountpoint."""

    name = "Icecast"

    class Options:
        """IcecastTrackObserver options."""

        def __init__(
            self,
            url: str,
            username: Optional[str] = None,
            password: Optional[str] = None,
            mount: Optional[str] = None,
        ):
            # TODO v3 remove optional args and only support parsed URLs
            (self.url, self.username, self.password, self.mount) = parse_icecast_url(
                url
            )
            # TODO v3 remove non URL usage of username, password, ...
            if not self.username and username:
                # grab from args if not in URL
                logger.warning("deprecated use username from URL")
                self.username = username
            if not self.username:
                # default to source if neither in URL nor args
                logger.warning("deprecated use username from URL")
                self.username = "source"
            if not self.password and password:
                # grab from args if not in URL
                logger.warning("deprecated use password from URL")
                self.password = password
            if not self.mount and mount:
                # grab from args if not in URL
                logger.warning("deprecated use mount from URL")
                self.mount = mount
            if not self.password:
                raise ValueError("Missing required parameter password for %s" % url)
            if not self.mount:
                raise ValueError("Missing required parameter mount for %s" % url)

    def __init__(self, options: Options):
        self.options = options
        logger.info(f"Icecast URL: {self.options.url} mount: {self.options.mount}")

    def track_started(self, track: Track):
        logger.info(
            "Updating Icecast Metadata for track: %s - %s" % (track.artist, track.title)
        )

        title = track.title

        if track.has_default_title() and track.has_default_artist():
            logger.info("Track has default info, using show instead")

            title = track.show.name

        params = {
            "mount": self.options.mount,
            "mode": "updinfo",
            "charset": "utf-8",
            "song": f"{track.artist} - {title}",
        }
        try:
            requests.get(
                self.options.url,
                auth=(self.options.username, self.options.password),
                params=params,
            )
        except requests.exceptions.RequestException as e:
            logger.exception(e)

        logger.info(
            f"Icecast Metadata updated on {self.options.url} with data: {params}"
        )

    def track_finished(self, track):
        return True



import logging
import urllib
from datetime import timedelta

import requests

from ..track import Track
from .base import TrackObserver

logger = logging.getLogger(__name__)


class DabAudioCompanionTrackObserver(TrackObserver):
    """Update track metadata in a DAB+ tranmission through the 'Audio Companion' API."""

    name = "DAB+ Audio Companion"

    def __init__(self, base_url, dls_enabled: bool = False):
        self.base_url = base_url + "/api/setDLS"
        self.dls_enabled = self.last_frame_was_dl_plus = dls_enabled
        logger.info(
            "DAB+ Audio Companion initialised with URL: %s, DLS+ enabled: %r"
            % (self.base_url, self.dls_enabled)
        )

    def track_started(self, track: Track):
        logger.info(
            "Updating DAB+ DLS for track: %s - %s" % (track.artist, track.title)
        )
        # TODO v3 remove _track_started_plain
        if not self.dls_enabled:
            return self._track_started_plain(track)

        params = {}

        if track.get_duration() < timedelta(seconds=5):
            logger.info(
                "Track is less than 5 seconds, not sending to DAB+ Audio Companion"
            )
            return

        if not track.has_default_title() and not track.has_default_artist():
            params["artist"] = track.artist
            params["title"] = track.title
            self.last_frame_was_dl_plus = True
        elif self.last_frame_was_dl_plus:
            logger.info(
                "Track has default info, using show instead. Sending DLS+ delete tags."
            )
            # track.artist contains station name if no artist is set
            message = f"{track.artist} - {track.show.name}"
            param = "".join(
                (
                    "##### parameters { #####\n",
                    "DL_PLUS=1\n",
                    # delete messages for artist and title as set by Audio Companion
                    f'DL_PLUS_TAG=1 {message.find(" ")} 0\n',
                    f'DL_PLUS_TAG=4 {message.find(" ")} 0\n',
                    "##### parameters } #####\n",
                    message,
                )
            )
            params["dls"] = param
            self.last_frame_was_dl_plus = False
        else:
            logger.info("Track has default info, using show instead")
            # track.artist contains station name if no artist is set
            params["dls"] = f"{track.artist} - {track.show.name}"

        logger.info(
            f"DAB+ Audio Companion URL: {self.base_url} data: {params} is DL+: {self.last_frame_was_dl_plus}"
        )

        resp = requests.post(self.base_url, params)
        if resp.status_code != 200:
            logger.error(f"DAB+ Audio Companion API call failed: {resp.text}")

    def _track_started_plain(self, track):
        # TODO v3 remove once we always send DLS with v3
        title = track.title

        if track.has_default_title() and track.has_default_artist():
            logger.info(
                "%s: Track has default info, using show instead" % self.__class__
            )

            title = track.show.name

        # artist is an unicode string which we have to encode into UTF-8
        # http://bugs.python.org/issue216716
        song_string = urllib.parse.quote_plus(
            "%s - %s" % (track.artist.encode("utf8"), title.encode("utf8"))
        )

        update_url = f"{self.base_url}?dls={song_string}"

        logger.info("DAB+ Audio Companion URL: " + update_url)

        urllib.request.urlopen(update_url)

    def track_finished(self, track):
        return True



# TODO scrobbling is not currently supported remove the no cover pragma from this file if you support it again
import calendar  # pragma: no cover
import logging  # pragma: no cover

import pylast  # pragma: no cover

from .base import TrackObserver  # pragma: no cover

logger = logging.getLogger(__name__)  # pragma: no cover


class ScrobblerTrackObserver(TrackObserver):  # pragma: no cover
    """Scrobble track information to last.fm and libre.fm."""

    name = "Audioscrobbler"

    lastfm_api_key = ""
    lastfm_api_secret = ""

    librefm_api_key = ""
    librefm_api_secret = ""

    def __init__(self, user, password):
        raise Exception("Unmaintained ScrobblerTrackObserver called")

        password = pylast.md5(password)

        self.lastfm = pylast.get_lastfm_network("", "", None, user, password)
        self.librefm = pylast.get_librefm_network("", "", None, user, password)

        self.scrobblers = {
            "lastfm": self.lastfm.get_scrobbler("tst", "1.0"),
            "librefm": self.librefm.get_scrobbler("tst", "1.0"),
        }

    def track_started(self, track):

        if track.has_default_artist() or track.has_default_title():
            logger.error("Default artist or title, skip now-playing scrobble")
            return False

        logger.info(
            "AS now-playing notification for track: %s - %s"
            % (track.artist, track.title)
        )

        # scrobble to all networks
        #
        # Note that track.get_duration() returns a timedelta object, which
        # could (theoretically) also contain minutes, hours and days, however
        # the track.set_duration() method only accepts seconds, therefore it's
        # safe to read track.get_duration().seconds. Python 2.7 will bring the
        # proper solution with its timedelta.total_seconds() method.
        for k, scrobbler in self.scrobblers.items():
            scrobbler.report_now_playing(
                track.artist,
                track.title,
                track.album,
                str(track.get_duration().seconds),
                str(track.track),
            )

    def track_finished(self, track):
        logger.info("Track duration was: %s" % track.get_duration().seconds)

        # if track.get_duration() < datetime.timedelta(seconds=30):
        if track.get_duration().seconds < 30:
            # respecting last.fm minimum song lenght policy of 30 seconds
            logger.info("Track duration is < 30s, skip scrobble submit")
            return False

        if track.has_default_artist() or track.has_default_title():
            logger.error("Default artist or title, skip scrobble submit")
            return False

        logger.info("AS submit for track: %s - %s" % (track.artist, track.title))

        # Get UTC unix timestamp without daylight saving informations
        # http://stackoverflow.com/questions/1077285/how-to-specify-time-zone-utc-when-converting-to-unix-time-python
        timestamp = calendar.timegm(track.starttime.utctimetuple())

        # scrobbling to all networks
        #
        # Note that track.get_duration() returns a timedelta object, which
        # could (theoretically) also contain minutes, hours and days, however
        # the track.set_duration() method only accepts seconds, therefore it's
        # safe to read track.get_duration().seconds. Python 2.7 will bring the
        # proper solution with its timedelta.total_seconds() method.
        for k, scrobbler in self.scrobblers.items():
            scrobbler.scrobble(
                track.artist,
                track.title,
                int(timestamp),
                "P",
                "",
                str(track.get_duration().seconds),
                track.album,
                str(track.track),
            )



import logging
import os
import signal
import sys
import time

from .input import observer as inputObservers
from .input.handler import InputHandler
from .misc.saemubox import SaemuBox
from .track.handler import TrackEventHandler
from .track.observers.dab_audio_companion import DabAudioCompanionTrackObserver
from .track.observers.icecast import IcecastTrackObserver
from .track.observers.ticker import TickerTrackObserver

logger = logging.getLogger(__name__)


class NowPlayingDaemon:
    """The daemon part of the nowplaying app runs at all time and coordinates the input and output."""

    """initialize last_input to a know value."""
    last_input = 1

    def __init__(self, options):
        self.options = options

        self.saemubox = SaemuBox(self.options.saemubox_ip)

    def main(self):  # pragma: no cover
        # TODO test once there is not saemubox in the loop
        logger.info("Starting up now-playing daemon")
        self.saemubox.run()

        try:
            self.register_signal_handlers()

            input_handler = self.get_input_handler()
        except Exception as e:
            logger.exception("Error: %s", e)
            sys.exit(-1)

        while True:
            try:
                saemubox_id = self.poll_saemubox()

                input_handler.update(saemubox_id)
            except Exception as e:
                logger.exception("Error: %s", e)

            time.sleep(self.options.sleepSeconds)

    def register_signal_handlers(self):
        logger.debug("Registering signal handler")
        signal.signal(signal.SIGINT, self.signal_handler)
        # signal.signal(signal.SIGKIL, self.signal_handler)

    def signal_handler(self, signum, frame):
        logger.debug("Signal %i caught" % signum)

        if signum == signal.SIGINT or signum == signal.SIGKILL:
            logger.info("Signal %i caught, terminating." % signum)
            sys.exit(os.EX_OK)

    def get_track_handler(self):  # pragma: no cover
        # TODO test once options have been refactored with v3
        handler = TrackEventHandler()
        [
            handler.register_observer(
                IcecastTrackObserver(
                    # TODO v3 remove uername and password because we mandate specifying via url
                    options=IcecastTrackObserver.Options(
                        url=url,
                        username="source",
                        password=self.options.icecastPassword,
                    )
                )
            )
            for url in self.options.icecast
        ]
        [
            handler.register_observer(
                DabAudioCompanionTrackObserver(
                    base_url=url, dls_enabled=self.options.dab_send_dls
                )
            )
            for url in self.options.dab
        ]
        handler.register_observer(TickerTrackObserver(self.options.tickerOutputFile))

        return handler

    def get_input_handler(self):  # pragma: no cover
        # TODO test once options have been refactored with v3
        handler = InputHandler()
        track_handler = self.get_track_handler()

        klangbecken = inputObservers.KlangbeckenInputObserver(
            self.options.currentShowUrl, self.options.inputFile
        )
        klangbecken.add_track_handler(track_handler)
        handler.register_observer(klangbecken)

        nonklangbecken = inputObservers.NonKlangbeckenInputObserver(
            self.options.currentShowUrl
        )
        nonklangbecken.add_track_handler(track_handler)
        handler.register_observer(nonklangbecken)

        return handler

    def poll_saemubox(self):  # pragma: no cover
        """
        Poll Saemubox for new data.

        Should be run once per main loop.

        TODO v3 remove once replaced with pathfinder
        """

        saemubox_id = self.saemubox.get_active_output_id()
        logger.debug("Sämubox id: %i" % saemubox_id)

        if self.last_input != saemubox_id:
            logger.info(
                'Sämubox changed from "%s" to "%s"'
                % (
                    self.saemubox.get_id_as_name(self.last_input),
                    self.saemubox.get_id_as_name(saemubox_id),
                )
            )

        self.last_input = saemubox_id
        return saemubox_id



"""OpenTelemetry for nowplaying.

This sets up our logging stack to use OpenTelemetry.
"""

import logging
import os
from datetime import datetime

from opentelemetry.exporter.otlp.proto.grpc._log_exporter import OTLPLogExporter
from opentelemetry.sdk._logs import (
    LogEmitterProvider,
    OTLPHandler,
    set_log_emitter_provider,
)
from opentelemetry.sdk._logs.export import (
    BatchLogProcessor,
    ConsoleLogExporter,
    SimpleLogProcessor,
)
from opentelemetry.sdk.resources import Resource


def _log_formatter(record):  # pragma: no cover
    return f"{datetime.fromtimestamp(record.timestamp/1000000000)} - {record.severity_text[:4]:4} - {record.attributes['source']['name'][11:]:14} - {record.body} - {record.attributes['source']['pathname']}:{record.attributes['source']['lineno']}\n"


class SourceAttributeFilter(logging.Filter):  # pragma: no cover
    """Used on the handler to ensure that some attributes are carried over to otel."""

    def filter(self, record) -> bool:
        record.source = {
            "name": record.name,
            "pathname": os.path.relpath(
                record.pathname, os.path.dirname(os.path.dirname(__file__))
            ),
            "lineno": record.lineno,
        }
        return True


def setup_otel(otlp_enable=False):  # pragma: no cover
    """Configure opentelemetry logging to stdout and collector."""
    root = logging.getLogger()
    root.setLevel(logging.INFO)

    log_emitter_provider = LogEmitterProvider(
        resource=Resource.create(
            {
                "service.name": "nowplaying",
            },
        )
    )
    set_log_emitter_provider(log_emitter_provider)

    console_exporter = ConsoleLogExporter(
        formatter=lambda record: _log_formatter(record)
    )
    log_emitter_provider.add_log_processor(SimpleLogProcessor(console_exporter))

    if otlp_enable:
        oltp_exporter = OTLPLogExporter(insecure=True)
        log_emitter_provider.add_log_processor(BatchLogProcessor(oltp_exporter))

    log_emitter = log_emitter_provider.get_log_emitter(__name__, "0.1")

    handler = OTLPHandler(level=logging.NOTSET, log_emitter=log_emitter)
    handler.addFilter(SourceAttributeFilter())

    root.addHandler(handler)

    logger = logging.getLogger(__name__)
    logger.info("Opentelemetry Configured (OTLP: %s)", otlp_enable)



import datetime
import logging
import uuid
import warnings

import isodate
import lxml.builder
import lxml.etree
import pytz

from .base import TrackObserver

logger = logging.getLogger(__name__)


class TickerTrackObserver(TrackObserver):
    """Writes the new ticker feed file.

    The feed file will be consumed by the song-ticker on the RaBe website. This is the
    successor of the long gone PseudoRssTrackObserver format used by the pre-WordPress
    website. This version here gets consumed by the WordPress website.
    """

    name = "Ticker"

    def __init__(self, ticker_file_path):
        warnings.warn(
            "The XML ticker format will be replaced with a JSON variant in the future",
            PendingDeprecationWarning,
        )
        self.ticker_file_path = ticker_file_path

    def track_started(self, track):
        logger.info(
            "Updating Ticker XML file for track: %s - %s" % (track.artist, track.title)
        )
        try:
            tz = pytz.timezone("Europe/Zurich")
        except pytz.exceptions.UnknownTimeZoneError:  # pragma: no coverage due to not knowing how to trigger
            tz = pytz.timezone("UTC")

        now = isodate.datetime_isoformat(datetime.datetime.now(tz))

        MAIN_NAMESPACE = "http://rabe.ch/schema/ticker.xsd"
        XLINK_NAMESPACE = "http://www.w3.org/1999/xlink"
        XLINK = "{%s}" % XLINK_NAMESPACE

        E = lxml.builder.ElementMaker(
            namespace=MAIN_NAMESPACE,
            nsmap={None: MAIN_NAMESPACE, "xlink": XLINK_NAMESPACE},
        )
        show_ref = E.link(track.show.url)
        show_ref.attrib[XLINK + "type"] = "simple"
        show_ref.attrib[XLINK + "href"] = track.show.url
        show_ref.attrib[XLINK + "show"] = "replace"

        ticker = E.ticker(
            E.identifier("ticker-%s" % uuid.uuid4()),
            E.creator("now-playing daemon v2"),
            E.date(now),
            E.show(
                E.name(track.show.name),
                show_ref,
                E.startTime(
                    isodate.datetime_isoformat(track.show.starttime.astimezone(tz))
                ),
                E.endTime(
                    isodate.datetime_isoformat(track.show.endtime.astimezone(tz))
                ),
                id=track.show.uuid,
            ),
            E.track(
                E.show(track.show.name, ref=track.show.uuid),
                E.artist(track.artist),
                E.title(track.title),
                E.startTime(isodate.datetime_isoformat(track.starttime.astimezone(tz))),
                E.endTime(isodate.datetime_isoformat(track.endtime.astimezone(tz))),
                id=track.uuid,
            ),
        )
        lxml.etree.ElementTree(ticker).write(
            self.ticker_file_path,
            pretty_print=True,
            xml_declaration=True,
            encoding="utf-8",
        )

    def track_finished(self, track):
        return True



import socket

from .daemon import NowPlayingDaemon
from .options import Options
from .otel import setup_otel


class NowPlaying:
    def run(self):
        """Load configuration, initialize environment and start nowplaying daemon."""
        self.options = Options()
        self.options.parse_known_args()

        self._setup_otel()
        socket.setdefaulttimeout(self.options.socketDefaultTimeout)
        self._run_daemon()

    def _setup_otel(self):  # pragma: no cover
        if not self.options.debug:
            setup_otel(self.options.otlp_enable)

    def _run_daemon(self):
        """Start nowplaying daemon."""
        NowPlayingDaemon(self.options).main()


if __name__ == "__main__":  # pragma: no cover
    NowPlaying().run()
